@@ -1,6 +1,6 @@
 -- |
 -- Module      :  $Header$
--- Copyright   :  (c) 2013-2015 Galois, Inc.
+-- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
 -- Stability   :  provisional
@@ -10,49 +10,56 @@
 -- patterns.  It also eliminates pattern bindings by de-sugaring them
 -- into `Bind`.  Furthermore, here we associate signatures and pragmas
 -- with the names to which they belong.
+{-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE DeriveGeneric #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE OverloadedStrings #-}
 module Cryptol.Parser.NoPat (RemovePatterns(..),Error(..)) where
 
-import Cryptol.Prims.Syntax
 import Cryptol.Parser.AST
-import Cryptol.Parser.Position(Range(..),start)
+import Cryptol.Parser.Position(Range(..),emptyRange,start,at)
+import Cryptol.Parser.Names (namesP)
 import Cryptol.Utils.PP
 import Cryptol.Utils.Panic(panic)
 
-import           MonadLib
+import           MonadLib hiding (mapM)
 import           Data.Maybe(maybeToList)
 import           Data.Either(partitionEithers)
 import qualified Data.Map as Map
 
-#if __GLASGOW_HASKELL__ < 710
-import           Control.Applicative(Applicative(..),(<$>))
-import           Data.Traversable(traverse)
-#endif
+import GHC.Generics (Generic)
+import Control.DeepSeq.Generics
+
+import Prelude ()
+import Prelude.Compat
 
 class RemovePatterns t where
   -- | Eliminate all patterns in a program.
   removePatterns :: t -> (t, [Error])
 
-instance RemovePatterns Program where
+instance RemovePatterns (Program PName) where
   removePatterns p = runNoPatM (noPatProg p)
 
-instance RemovePatterns Expr where
+instance RemovePatterns (Expr PName) where
   removePatterns e = runNoPatM (noPatE e)
 
-instance RemovePatterns Module where
+instance RemovePatterns (Module PName) where
   removePatterns m = runNoPatM (noPatModule m)
 
-instance RemovePatterns [Decl] where
+instance RemovePatterns [Decl PName] where
   removePatterns ds = runNoPatM (noPatDs ds)
 
-simpleBind :: Located QName -> Expr -> Bind
-simpleBind x e = Bind { bName = x, bParams = [], bDef = e
+simpleBind :: Located PName -> Expr PName -> Bind PName
+simpleBind x e = Bind { bName = x, bParams = []
+                      , bDef = at e (Located emptyRange (DExpr e))
                       , bSignature = Nothing, bPragmas = []
-                      , bMono = True
+                      , bMono = True, bInfix = False, bFixity = Nothing
+                      , bDoc = Nothing
                       }
 
-sel :: Pattern -> QName -> Selector -> Bind
+sel :: Pattern PName -> PName -> Selector -> Bind PName
 sel p x s = let (a,ts) = splitSimpleP p
             in simpleBind a (foldl ETyped (ESel (EVar x) s) ts)
 
@@ -60,7 +67,7 @@ sel p x s = let (a,ts) = splitSimpleP p
 -- Simple patterns may only contain variables and type annotations.
 
 -- XXX: We can replace the types in the selcetors with annotations on the bindings.
-noPat :: Pattern -> NoPatM (Pattern, [Bind])
+noPat :: Pattern PName -> NoPatM (Pattern PName, [Bind PName])
 noPat pat =
   case pat of
     PVar x -> return (PVar x, [])
@@ -74,10 +81,9 @@ noPat pat =
       do (as,dss) <- unzip `fmap` mapM noPat ps
          x <- newName
          r <- getRange
-         let qx       = mkUnqual x
          let len      = length ps
              ty       = TTuple (replicate len TWild)
-             getN a n = sel a qx (TupleSel n (Just len))
+             getN a n = sel a x (TupleSel n (Just len))
          return (pTy r x ty, zipWith getN as [0..] ++ concat dss)
 
     PList [] ->
@@ -89,20 +95,18 @@ noPat pat =
       do (as,dss) <- unzip `fmap` mapM noPat ps
          x <- newName
          r <- getRange
-         let qx       = mkUnqual x
-             len      = length ps
+         let len      = length ps
              ty       = TSeq (TNum (fromIntegral len)) TWild
-             getN a n = sel a qx (ListSel n (Just len))
+             getN a n = sel a x (ListSel n (Just len))
          return (pTy r x ty, zipWith getN as [0..] ++ concat dss)
 
     PRecord fs ->
       do (as,dss) <- unzip `fmap` mapM (noPat . value) fs
          x <- newName
          r <- getRange
-         let qx       = mkUnqual x
-             shape    = map (thing . name) fs
+         let shape    = map (thing . name) fs
              ty       = TRecord (map (fmap (\_ -> TWild)) fs)
-             getN a n = sel a qx (RecordSel n (Just shape))
+             getN a n = sel a x (RecordSel n (Just shape))
          return (pTy r x ty, zipWith getN as shape ++ concat dss)
 
     PTyped p t ->
@@ -116,11 +120,10 @@ noPat pat =
          x <- newName
          tmp <- newName
          r <- getRange
-         let qx   = mkUnqual x
-             qtmp = mkUnqual tmp
-             bTmp = simpleBind (Located r qtmp) (EApp (ECon ECSplitAt) (EVar qx))
-             b1   = sel a1 qtmp (TupleSel 0 (Just 2))
-             b2   = sel a2 qtmp (TupleSel 1 (Just 2))
+         let prim = EVar (mkUnqual (mkIdent "splitAt"))
+             bTmp = simpleBind (Located r tmp) (EApp prim (EVar x))
+             b1   = sel a1 tmp (TupleSel 0 (Just 2))
+             b2   = sel a2 tmp (TupleSel 1 (Just 2))
          return (pVar r x, bTmp : b1 : b2 : ds1 ++ ds2)
 
     PLocated p r1 -> inRange r1 (noPat p)
@@ -130,8 +133,8 @@ noPat pat =
   pTy  r x t = PTyped (PVar (Located r x)) t
 
 
-splitSimpleP :: Pattern -> (Located QName, [Type])
-splitSimpleP (PVar x)     = (fmap mkUnqual x, [])
+splitSimpleP :: Pattern PName -> (Located PName, [Type PName])
+splitSimpleP (PVar x)     = (x, [])
 splitSimpleP (PTyped p t) = let (x,ts) = splitSimpleP p
                             in (x, t:ts)
 splitSimpleP p            = panic "splitSimpleP"
@@ -139,11 +142,10 @@ splitSimpleP p            = panic "splitSimpleP"
 
 --------------------------------------------------------------------------------
 
-noPatE :: Expr -> NoPatM Expr
+noPatE :: Expr PName -> NoPatM (Expr PName)
 noPatE expr =
   case expr of
     EVar {}       -> return expr
-    ECon {}       -> return expr
     ELit {}       -> return expr
     ETuple es     -> ETuple  <$> mapM noPatE es
     ERecord es    -> ERecord <$> mapM noPatF es
@@ -162,10 +164,13 @@ noPatE expr =
                         return (EFun ps1 e1)
     ELocated e r1 -> ELocated <$> inRange r1 (noPatE e) <*> return r1
 
+    EParens e     -> EParens <$> noPatE e
+    EInfix x y f z-> EInfix  <$> noPatE x <*> pure y <*> pure f <*> noPatE z
+
   where noPatF x = do e <- noPatE (value x)
                       return x { value = e }
 
-noPatFun :: [Pattern] -> Expr -> NoPatM ([Pattern], Expr)
+noPatFun :: [Pattern PName] -> Expr PName -> NoPatM ([Pattern PName], Expr PName)
 noPatFun ps e =
   do (xs,bs) <- unzip <$> mapM noPat ps
      e1 <- noPatE e
@@ -175,26 +180,34 @@ noPatFun ps e =
      return (xs, body)
 
 
-noPatArm :: [Match] -> NoPatM [Match]
+noPatArm :: [Match PName] -> NoPatM [Match PName]
 noPatArm ms = concat <$> mapM noPatM ms
 
-noPatM :: Match -> NoPatM [Match]
+noPatM :: Match PName -> NoPatM [Match PName]
 noPatM (Match p e) =
   do (x,bs) <- noPat p
      e1     <- noPatE e
      return (Match x e1 : map MatchLet bs)
 noPatM (MatchLet b) = (return . MatchLet) <$> noMatchB b
 
-noMatchB :: Bind -> NoPatM Bind
+noMatchB :: Bind PName -> NoPatM (Bind PName)
 noMatchB b =
-  do (ps,e) <- noPatFun (bParams b) (bDef b)
-     return b { bParams = ps, bDef = e }
+  case thing (bDef b) of
 
-noMatchD :: Decl -> NoPatM [Decl]
+    DPrim | null (bParams b) -> return b
+          | otherwise        -> panic "NoPat" [ "noMatchB: primitive with params"
+                                              , show b ]
+
+    DExpr e ->
+      do (ps,e') <- noPatFun (bParams b) e
+         return b { bParams = ps, bDef = DExpr e' <$ bDef b }
+
+noMatchD :: Decl PName -> NoPatM [Decl PName]
 noMatchD decl =
   case decl of
     DSignature {}   -> return [decl]
     DPragma {}      -> return [decl]
+    DFixity{}       -> return [decl]
 
     DBind b         -> do b1 <- noMatchB b
                           return [DBind b1]
@@ -205,22 +218,27 @@ noMatchD decl =
                           let e2 = foldl ETyped e1 ts
                           return $ DBind Bind { bName = x
                                               , bParams = []
-                                              , bDef = e2
+                                              , bDef = at e (Located emptyRange (DExpr e2))
                                               , bSignature = Nothing
                                               , bPragmas = []
                                               , bMono = False
+                                              , bInfix = False
+                                              , bFixity = Nothing
+                                              , bDoc = Nothing
                                               } : map DBind bs
     DType {}        -> return [decl]
 
     DLocated d r1   -> do bs <- inRange r1 $ noMatchD d
                           return $ map (`DLocated` r1) bs
 
-noPatDs :: [Decl] -> NoPatM [Decl]
+noPatDs :: [Decl PName] -> NoPatM [Decl PName]
 noPatDs ds =
   do ds1 <- concat <$> mapM noMatchD ds
      let pragmaMap = Map.fromListWith (++) $ concatMap toPragma ds1
          sigMap    = Map.fromListWith (++) $ concatMap toSig ds1
-     (ds2, (pMap,sMap)) <- runStateT (pragmaMap, sigMap) $ annotDs ds1
+         fixMap    = Map.fromListWith (++) $ concatMap toFixity ds1
+     (ds2, (pMap,sMap,fMap,_)) <- runStateT (pragmaMap, sigMap, fixMap, Map.empty)
+                                            (annotDs ds1)
 
      forM_ (Map.toList pMap) $ \(n,ps) ->
        forM_ ps $ \p -> recordError $ PragmaNoBind (p { thing = n }) (thing p)
@@ -230,18 +248,24 @@ noPatDs ds =
           forM_ ss $ \s -> recordError $ SignatureNoBind (s { thing = n })
                                                          (thing s)
 
+     forM_ (Map.toList fMap) $ \(n,fs) ->
+       forM_ fs $ \f -> recordError $ FixityNoBind f { thing = n }
+
      return ds2
 
-noPatTopDs :: [TopLevel Decl] -> NoPatM [TopLevel Decl]
+noPatTopDs :: [TopLevel (Decl PName)] -> NoPatM [TopLevel (Decl PName)]
 noPatTopDs tds =
   do noPatGroups <- mapM (noMatchD . tlValue) tds
 
      let allDecls  = concat noPatGroups
          pragmaMap = Map.fromListWith (++) $ concatMap toPragma allDecls
          sigMap    = Map.fromListWith (++) $ concatMap toSig    allDecls
+         fixMap    = Map.fromListWith (++) $ concatMap toFixity allDecls
+         docMap    = Map.fromListWith (++) $ concatMap toDocs   tds
 
      let exportGroups = zipWith (\ td ds -> td { tlValue = ds }) tds noPatGroups
-     (tds', (pMap,sMap)) <- runStateT (pragmaMap,sigMap) (annotTopDs exportGroups)
+     (tds', (pMap,sMap,fMap,_)) <- runStateT (pragmaMap,sigMap,fixMap,docMap)
+                                             (annotTopDs exportGroups)
 
      forM_ (Map.toList pMap) $ \(n,ps) ->
        forM_ ps $ \p -> recordError $ PragmaNoBind (p { thing = n }) (thing p)
@@ -251,10 +275,13 @@ noPatTopDs tds =
           forM_ ss $ \s -> recordError $ SignatureNoBind (s { thing = n })
                                                          (thing s)
 
+     forM_ (Map.toList fMap) $ \(n,fs) ->
+       forM_ fs $ \f -> recordError $ FixityNoBind f { thing = n }
+
      return tds'
 
 
-noPatProg :: Program -> NoPatM Program
+noPatProg :: Program PName -> NoPatM (Program PName)
 noPatProg (Program topDs) =
   do let (ds, others) = partitionEithers (map isDecl topDs)
      ds1 <- noPatTopDs ds
@@ -264,7 +291,7 @@ noPatProg (Program topDs) =
   isDecl (Decl d) = Left d
   isDecl d        = Right d
 
-noPatModule :: Module -> NoPatM Module
+noPatModule :: Module PName -> NoPatM (Module PName)
 noPatModule m =
   do let (ds, others) = partitionEithers (map isDecl (mDecls m))
      ds1 <- noPatTopDs ds
@@ -277,8 +304,10 @@ noPatModule m =
 
 --------------------------------------------------------------------------------
 
-type AnnotMap = ( Map.Map QName [Located Pragma]
-                , Map.Map QName [Located Schema]
+type AnnotMap = ( Map.Map PName [Located  Pragma       ]
+                , Map.Map PName [Located (Schema PName)]
+                , Map.Map PName [Located  Fixity       ]
+                , Map.Map PName [Located  String       ]
                 )
 
 -- | Add annotations to exported declaration groups.
@@ -287,7 +316,8 @@ type AnnotMap = ( Map.Map QName [Located Pragma]
 -- export specifications, this will favor the specification of the binding.
 -- This is most likely the intended behavior, so it's probably fine, but it does
 -- smell a bit.
-annotTopDs :: [TopLevel [Decl]] -> StateT AnnotMap NoPatM [TopLevel Decl]
+annotTopDs :: [TopLevel [Decl PName]]
+           -> StateT AnnotMap NoPatM [TopLevel (Decl PName)]
 annotTopDs tds =
   case tds of
 
@@ -302,7 +332,7 @@ annotTopDs tds =
 
 
 -- | Add annotations, keeping track of which annotation are not yet used up.
-annotDs :: [Decl] -> StateT AnnotMap NoPatM [Decl]
+annotDs :: [Decl PName] -> StateT AnnotMap NoPatM [Decl PName]
 annotDs (d : ds) =
   do ignore <- runExceptionT (annotD d)
      case ignore of
@@ -312,54 +342,96 @@ annotDs [] = return []
 
 -- | Add annotations, keeping track of which annotation are not yet used up.
 -- The exception indicates which declarations are no longer needed.
-annotD :: Decl -> ExceptionT () (StateT AnnotMap NoPatM) Decl
+annotD :: Decl PName -> ExceptionT () (StateT AnnotMap NoPatM) (Decl PName)
 annotD decl =
   case decl of
     DBind b       -> DBind <$> lift (annotB b)
     DSignature {} -> raise ()
+    DFixity{}     -> raise ()
     DPragma {}    -> raise ()
     DPatBind {}   -> raise ()
     DType {}      -> return decl
     DLocated d r  -> (`DLocated` r) <$> annotD d
 
 -- | Add pragma/signature annotations to a binding.
-annotB :: Bind -> StateT AnnotMap NoPatM Bind
+annotB :: Bind PName -> StateT AnnotMap NoPatM (Bind PName)
 annotB Bind { .. } =
-  do (ps,ss) <- get
-     let name = thing bName
-     case ( Map.updateLookupWithKey (\_ _ -> Nothing) name ps
-          , Map.updateLookupWithKey (\_ _ -> Nothing) name ss
+  do (ps,ss,fs,ds) <- get
+     let name       = thing bName
+         remove _ _ = Nothing
+     case ( Map.updateLookupWithKey remove name ps
+          , Map.updateLookupWithKey remove name ss
+          , Map.updateLookupWithKey remove name fs
+          , Map.updateLookupWithKey remove name ds
           ) of
-           ( (thisPs, pragmas1) , (thisSigs, sigs1)) ->
+           ( (thisPs, pragmas1), (thisSigs, sigs1), (thisFixes, fixes1), (thisDocs, docs1)) ->
                 do s <- lift $ checkSigs name (jn thisSigs)
-                   set (pragmas1,sigs1)
+                   f <- lift $ checkFixs name (jn thisFixes)
+                   d <- lift $ checkDocs name (jn thisDocs)
+                   set (pragmas1,sigs1,fixes1,docs1)
                    return Bind { bSignature = s
                                , bPragmas = map thing (jn thisPs) ++ bPragmas
+                               , bFixity = f
+                               , bDoc = d
                                , ..
                                }
   where jn x = concat (maybeToList x)
 
 -- | Check for multiple signatures.
-checkSigs :: QName -> [Located Schema] -> NoPatM (Maybe Schema)
+checkSigs :: PName -> [Located (Schema PName)] -> NoPatM (Maybe (Schema PName))
 checkSigs _ []             = return Nothing
 checkSigs _ [s]            = return (Just (thing s))
 checkSigs f xs@(s : _ : _) = do recordError $ MultipleSignatures f xs
                                 return (Just (thing s))
 
+checkFixs :: PName -> [Located Fixity] -> NoPatM (Maybe Fixity)
+checkFixs _ []       = return Nothing
+checkFixs _ [f]      = return (Just (thing f))
+checkFixs f fs@(x:_) = do recordError $ MultipleFixities f $ map srcRange fs
+                          return (Just (thing x))
+
+
+checkDocs :: PName -> [Located String] -> NoPatM (Maybe String)
+checkDocs _ []       = return Nothing
+checkDocs _ [d]      = return (Just (thing d))
+checkDocs f ds@(d:_) = do recordError $ MultipleDocs f (map srcRange ds)
+                          return (Just (thing d))
+
 
 -- | Does this declaration provide some signatures?
-toSig :: Decl -> [(QName, [Located Schema])]
+toSig :: Decl PName -> [(PName, [Located (Schema PName)])]
 toSig (DLocated d _)      = toSig d
 toSig (DSignature xs s)   = [ (thing x,[Located (srcRange x) s]) | x <- xs ]
 toSig _                   = []
 
 -- | Does this declaration provide some signatures?
-toPragma :: Decl -> [(QName, [Located Pragma])]
+toPragma :: Decl PName -> [(PName, [Located Pragma])]
 toPragma (DLocated d _)   = toPragma d
 toPragma (DPragma xs s)   = [ (thing x,[Located (srcRange x) s]) | x <- xs ]
 toPragma _                = []
 
+-- | Does this declaration provide fixity information?
+toFixity :: Decl PName -> [(PName, [Located Fixity])]
+toFixity (DFixity f ns) = [ (thing n, [Located (srcRange n) f]) | n <- ns ]
+toFixity _              = []
 
+-- | Does this top-level declaration provide a documentation string?
+toDocs :: TopLevel (Decl PName) -> [(PName, [Located String])]
+toDocs TopLevel { .. }
+  | Just txt <- tlDoc = go txt tlValue
+  | otherwise = []
+  where
+  go txt decl =
+    case decl of
+      DSignature ns _ -> [ (thing n, [txt]) | n <- ns ]
+      DFixity _ ns    -> [ (thing n, [txt]) | n <- ns ]
+      DBind b         -> [ (thing (bName b), [txt]) ]
+      DLocated d _    -> go txt d
+      DPatBind p _    -> [ (thing n, [txt]) | n <- namesP p ]
+
+      -- XXX revisit these
+      DPragma _ _     -> []
+      DType _         -> []
 
 
 --------------------------------------------------------------------------------
@@ -367,10 +439,15 @@ newtype NoPatM a = M { unM :: ReaderT Range (StateT RW Id) a }
 
 data RW     = RW { names :: !Int, errors :: [Error] }
 
-data Error  = MultipleSignatures QName [Located Schema]
-            | SignatureNoBind (Located QName) Schema
-            | PragmaNoBind (Located QName) Pragma
-              deriving (Show)
+data Error  = MultipleSignatures PName [Located (Schema PName)]
+            | SignatureNoBind (Located PName) (Schema PName)
+            | PragmaNoBind (Located PName) Pragma
+            | MultipleFixities PName [Range]
+            | FixityNoBind (Located PName)
+            | MultipleDocs PName [Range]
+              deriving (Show,Generic)
+
+instance NFData Error where rnf = genericRnf
 
 instance Functor NoPatM where fmap = liftM
 instance Applicative NoPatM where pure = return; (<*>) = ap
@@ -381,7 +458,7 @@ instance Monad NoPatM where
 
 
 -- | Pick a new name, to be used when desugaring patterns.
-newName :: NoPatM Name
+newName :: NoPatM PName
 newName = M $ sets $ \s -> let x = names s
                            in (NewName NoPat x, s { names = x + 1 })
 
@@ -424,5 +501,16 @@ instance PP Error where
         text "Pragma without a matching binding:"
          $$ nest 2 (pp s)
 
+      MultipleFixities n locs ->
+        text "Multiple fixity declarations for" <+> quotes (pp n)
+        $$ nest 2 (vcat (map pp locs))
 
+      FixityNoBind n ->
+        text "At" <+> pp (srcRange n) <> colon <+>
+        text "Fixity declaration without a matching binding for:" <+>
+         pp (thing n)
+
+      MultipleDocs n locs ->
+        text "Multiple documentation blocks given for:" <+> pp n
+        $$ nest 2 (vcat (map pp locs))
 
