@@ -1,17 +1,12 @@
 -- |
 -- Module      :  $Header$
--- Copyright   :  (c) 2013-2015 Galois, Inc.
+-- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
 -- Stability   :  provisional
 -- Portability :  portable
 
-{-# LANGUAGE CPP #-}
-#if __GLASGOW_HASKELL__ >= 706
 {-# LANGUAGE RecursiveDo #-}
-#else
-{-# LANGUAGE DoRec, RecursiveDo #-}
-#endif
 {-# LANGUAGE Safe #-}
 module Cryptol.TypeCheck.Kind
   ( checkType
@@ -26,8 +21,9 @@ import           Cryptol.Parser.Position
 import           Cryptol.TypeCheck.AST
 import           Cryptol.TypeCheck.Monad hiding (withTParams)
 import           Cryptol.TypeCheck.Solve (simplifyAllConstraints
-                                         ,checkTypeFunction)
+                                         ,wfTypeFunction)
 import           Cryptol.Utils.PP
+import           Cryptol.Utils.Panic (panic)
 
 import qualified Data.Map as Map
 import           Data.List(sortBy,groupBy)
@@ -38,7 +34,7 @@ import           Control.Monad(unless,forM)
 
 
 -- | Check a type signature.
-checkSchema :: P.Schema -> InferM (Schema, [Goal])
+checkSchema :: P.Schema Name -> InferM (Schema, [Goal])
 checkSchema (P.Forall xs ps t mb) =
   do ((xs1,(ps1,t1)), gs) <-
         collectGoals $
@@ -54,7 +50,7 @@ checkSchema (P.Forall xs ps t mb) =
           Just r  -> inRange r
 
 -- | Check a type-synonym declaration.
-checkTySyn :: P.TySyn -> InferM TySyn
+checkTySyn :: P.TySyn Name -> InferM TySyn
 checkTySyn (P.TySyn x as t) =
   do ((as1,t1),gs) <- collectGoals
                     $ inRange (srcRange x)
@@ -69,7 +65,7 @@ checkTySyn (P.TySyn x as t) =
 
 -- | Check a newtype declaration.
 -- XXX: Do something with constraints.
-checkNewtype :: P.Newtype -> InferM Newtype
+checkNewtype :: P.Newtype Name -> InferM Newtype
 checkNewtype (P.Newtype x as fs) =
   do ((as1,fs1),gs) <- collectGoals $
        inRange (srcRange x) $
@@ -90,7 +86,7 @@ checkNewtype (P.Newtype x as fs) =
 
 
 
-checkType :: P.Type -> Maybe Kind -> InferM Type
+checkType :: P.Type Name -> Maybe Kind -> InferM Type
 checkType t k =
   do (_, t1) <- withTParams True [] $ doCheckType t k
      return t1
@@ -124,7 +120,7 @@ There are two reasons for this choice:
      annotation) in the rest.
 -}
 
-withTParams :: Bool -> [P.TParam] -> KindM a -> InferM ([TParam], a)
+withTParams :: Bool -> [P.TParam Name] -> KindM a -> InferM ([TParam], a)
 withTParams allowWildCards xs m =
   mdo mapM_ recordError duplicates
       (a, vars) <- runKindM allowWildCards (zip' xs ts) m
@@ -132,13 +128,13 @@ withTParams allowWildCards xs m =
       return (as,a)
   where
   getKind vs tp =
-    case Map.lookup (P.tpQName tp) vs of
+    case Map.lookup (P.tpName tp) vs of
       Just k  -> return k
       Nothing -> do recordWarning (DefaultingKind tp P.KNum)
                     return KNum
 
   newTP vs tp = do k <- getKind vs tp
-                   n <- newTParam (Just (mkUnqual (P.tpName tp))) k
+                   n <- newTParam (Just (P.tpName tp)) k
                    return (n, TVar (tpVar n))
 
 
@@ -146,7 +142,7 @@ withTParams allowWildCards xs m =
   This is needed to make the monadic recursion work correctly,
   because the data dependency is only on the part that is known. -}
   zip' [] _           = []
-  zip' (a:as) ~(t:ts) = (mkUnqual (P.tpName a), fmap cvtK (P.tpKind a), t) : zip' as ts
+  zip' (a:as) ~(t:ts) = (P.tpName a, fmap cvtK (P.tpKind a), t) : zip' as ts
 
   cvtK P.KNum  = KNum
   cvtK P.KType = KType
@@ -157,7 +153,7 @@ withTParams allowWildCards xs m =
 
 -- | Check an application of a type constant.
 tcon :: TCon            -- ^ Type constant being applied
-     -> [P.Type]        -- ^ Type parameters
+     -> [P.Type Name]   -- ^ Type parameters
      -> Maybe Kind      -- ^ Expected kind
      -> KindM Type      -- ^ Resulting type
 tcon tc ts0 k =
@@ -166,8 +162,8 @@ tcon tc ts0 k =
 
 -- | Check a use of a type-synonym, newtype, or scoped-type variable.
 tySyn :: Bool         -- ^ Should we check for scoped type vars.
-      -> QName        -- ^ Name of type sysnonym
-      -> [P.Type]     -- ^ Type synonym parameters
+      -> Name         -- ^ Name of type sysnonym
+      -> [P.Type Name]-- ^ Type synonym parameters
       -> Maybe Kind   -- ^ Expected kind
       -> KindM Type   -- ^ Resulting type
 tySyn scoped x ts k =
@@ -214,7 +210,7 @@ tySyn scoped x ts k =
 
 
 -- | Check a type-application.
-appTy :: [P.Type]             -- ^ Parameters to type function
+appTy :: [P.Type Name]        -- ^ Parameters to type function
       -> Kind                 -- ^ Kind of type function
       -> KindM ([Type], Kind) -- ^ Validated parameters, resulting kind
 appTy [] k1 = return ([],k1)
@@ -230,7 +226,7 @@ appTy ts k1 =
 
 
 -- | Validate a parsed type.
-doCheckType :: P.Type         -- ^ Type that needs to be checked
+doCheckType :: P.Type Name  -- ^ Type that needs to be checked
           -> Maybe Kind     -- ^ Expected kind (if any)
           -> KindM Type     -- ^ Checked type
 doCheckType ty k =
@@ -257,7 +253,7 @@ doCheckType ty k =
          -- constraints.
          case it of
            TCon (TF f) ts' ->
-              case checkTypeFunction f ts' of
+              case wfTypeFunction f ts' of
                  [] -> return ()
                  ps -> kNewGoals (CtPartialTypeFun (BuiltInTyFun f)) ps
            _ -> return ()
@@ -272,6 +268,12 @@ doCheckType ty k =
     P.TUser x []    -> checkTyThing x k
     P.TUser x ts    -> tySyn False x ts k
 
+    P.TParens t     -> doCheckType t k
+
+    P.TInfix{}      -> panic "KindCheck"
+                       [ "TInfix not removed by the renamer", show ty ]
+
+
   where
   checkF f = do t <- kInRange (srcRange (name f))
                    $ doCheckType (value f) (Just KType)
@@ -280,7 +282,7 @@ doCheckType ty k =
 
 
 -- | Check a type-variable or type-synonym.
-checkTyThing :: QName         -- ^ Name of thing that needs checking
+checkTyThing :: Name          -- ^ Name of thing that needs checking
              -> Maybe Kind    -- ^ Expected kind
              -> KindM Type
 checkTyThing x k =
@@ -299,7 +301,7 @@ checkTyThing x k =
 
 
 -- | Validate a parsed proposition.
-checkProp :: P.Prop           -- ^ Proposition that need to be checked
+checkProp :: P.Prop Name      -- ^ Proposition that need to be checked
           -> KindM Type       -- ^ Checked representation
 checkProp prop =
   case prop of
@@ -309,6 +311,7 @@ checkProp prop =
     P.CArith t1     -> tcon (PC PArith)         [t1]    (Just KProp)
     P.CCmp t1       -> tcon (PC PCmp)           [t1]    (Just KProp)
     P.CLocated p r1 -> kInRange r1 (checkProp p)
+    P.CType _       -> panic "checkProp" [ "Unexpected CType", show prop ]
 
 
 -- | Check that a type has the expected kind.
