@@ -1,54 +1,64 @@
 -- |
 -- Module      :  $Header$
--- Copyright   :  (c) 2013-2015 Galois, Inc.
+-- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
 -- Stability   :  provisional
 -- Portability :  portable
-
-{-# LANGUAGE RecordWildCards, CPP, Safe #-}
-#if __GLASGOW_HASKELL__ >= 706
+{-# LANGUAGE RecordWildCards, Safe #-}
 {-# LANGUAGE RecursiveDo #-}
-#else
-{-# LANGUAGE DoRec #-}
-#endif
+{-# LANGUAGE DeriveGeneric #-}
 module Cryptol.TypeCheck.Monad
   ( module Cryptol.TypeCheck.Monad
   , module Cryptol.TypeCheck.InferTypes
   ) where
 
+import           Cryptol.ModuleSystem.Name (SupplyT,runSupplyT,FreshM(..),Supply)
 import           Cryptol.Parser.Position
 import qualified Cryptol.Parser.AST as P
 import           Cryptol.TypeCheck.AST
 import           Cryptol.TypeCheck.Subst
 import           Cryptol.TypeCheck.Unify(mgu, Result(..), UnificationError(..))
 import           Cryptol.TypeCheck.InferTypes
+import qualified Cryptol.TypeCheck.Solver.CrySAT as CrySAT
 import           Cryptol.Utils.PP(pp, (<+>), Doc, text, quotes)
 import           Cryptol.Utils.Panic(panic)
 
+import qualified Control.Applicative as A
+import           Control.Monad.Fix(MonadFix(..))
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 import           Data.Map (Map)
 import           Data.Set (Set)
-import           Data.List(find)
+import           Data.List(find, minimumBy, groupBy, sortBy)
 import           Data.Maybe(mapMaybe)
-import           MonadLib
-import qualified Control.Applicative as A
-import           Control.Monad.Fix(MonadFix(..))
+import           Data.Function(on)
+import           MonadLib hiding (mapM)
+
+import GHC.Generics (Generic)
+import Control.DeepSeq.Generics
+
+import Prelude ()
+import Prelude.Compat
 
-#if __GLASGOW_HASKELL__ < 710
-import           Data.Functor
-#endif
 
 -- | Information needed for type inference.
 data InferInput = InferInput
   { inpRange     :: Range             -- ^ Location of program source
-  , inpVars      :: Map QName Schema  -- ^ Variables that are in scope
-  , inpTSyns     :: Map QName TySyn   -- ^ Type synonyms that are in scope
-  , inpNewtypes  :: Map QName Newtype -- ^ Newtypes in scope
+  , inpVars      :: Map Name Schema   -- ^ Variables that are in scope
+  , inpTSyns     :: Map Name TySyn    -- ^ Type synonyms that are in scope
+  , inpNewtypes  :: Map Name Newtype  -- ^ Newtypes in scope
   , inpNameSeeds :: NameSeeds         -- ^ Private state of type-checker
   , inpMonoBinds :: Bool              -- ^ Should local bindings without
                                       --   signatures be monomorphized?
+
+  , inpSolverConfig :: SolverConfig   -- ^ Options for the constraint solver
+
+  , inpPrimNames :: !PrimMap          -- ^ The mapping from 'Ident' to 'Name',
+                                      -- for names that the typechecker
+                                      -- needs to refer to.
+
+  , inpSupply :: !Supply              -- ^ The supply for fresh name generation
   } deriving Show
 
 
@@ -56,7 +66,9 @@ data InferInput = InferInput
 data NameSeeds = NameSeeds
   { seedTVar    :: !Int
   , seedGoal    :: !Int
-  } deriving Show
+  } deriving (Show, Generic)
+
+instance NFData NameSeeds where rnf = genericRnf
 
 -- | The initial seeds, used when checking a fresh program.
 nameSeeds :: NameSeeds
@@ -68,14 +80,14 @@ data InferOutput a
   = InferFailed [(Range,Warning)] [(Range,Error)]
     -- ^ We found some errors
 
-  | InferOK [(Range,Warning)] NameSeeds a
+  | InferOK [(Range,Warning)] NameSeeds Supply a
     -- ^ Type inference was successful.
 
 
     deriving Show
 
 runInferM :: TVars a => InferInput -> InferM a -> IO (InferOutput a)
-runInferM info (IM m) =
+runInferM info (IM m) = CrySAT.withSolver (inpSolverConfig info) $ \solver ->
   do rec ro <- return RO { iRange     = inpRange info
                      , iVars          = Map.map ExtVar (inpVars info)
                      , iTVars         = []
@@ -83,9 +95,13 @@ runInferM info (IM m) =
                      , iNewtypes      = fmap mkExternal (inpNewtypes info)
                      , iSolvedHasLazy = iSolvedHas finalRW     -- RECURSION
                      , iMonoBinds     = inpMonoBinds info
+                     , iSolver        = solver
+                     , iPrimNames     = inpPrimNames info
                      }
 
-         (result, finalRW) <- runStateT rw $ runReaderT ro m  -- RECURSION
+         ((result, finalRW),supply') <-
+             runSupplyT (inpSupply info) $ runStateT rw
+                                         $ runReaderT ro m  -- RECURSION
 
      let theSu    = iSubst finalRW
          defSu    = defaultingSubst theSu
@@ -98,13 +114,17 @@ runInferM info (IM m) =
              | nullGoals cts
                    -> return $ InferOK warns
                                   (iNameSeeds finalRW)
+                                  supply'
                                   (apSubst defSu result)
            (cts,has) -> return $ InferFailed warns
+                $ dropErrorsFromSameLoc
                 [ ( goalRange g
-                  , UnsolvedGoal (apSubst theSu g)
+                  , UnsolvedGoal False (apSubst theSu g)
                   ) | g <- fromGoals cts ++ map hasGoal has
                 ]
-       errs -> return $ InferFailed warns [(r,apSubst theSu e) | (r,e) <- errs]
+       errs -> return $ InferFailed warns
+                      $ dropErrorsFromSameLoc
+                                  [(r,apSubst theSu e) | (r,e) <- errs]
 
   where
   mkExternal x = (IsExternal, x)
@@ -120,23 +140,34 @@ runInferM info (IM m) =
           , iSolvedHas  = Map.empty
           }
 
+  dropErrorsFromSameLoc = map chooseBestError . groupBy ((==)    `on` fst)
+                                              . sortBy  (cmpRange `on` fst)
 
-newtype InferM a = IM { unIM :: ReaderT RO (StateT RW IO) a }
+  addErrorSize (r,e) = (length (show (pp e)), (r,e))
+  chooseBestError    = snd . minimumBy (compare `on` fst) . map addErrorSize
+
+  -- The actual order does not matter
+  cmpRange (Range x y z) (Range a b c) = compare (x,y,z) (a,b,c)
+
+
+
+
+newtype InferM a = IM { unIM :: ReaderT RO (StateT RW (SupplyT IO)) a }
 
 data DefLoc = IsLocal | IsExternal
 
 -- | Read-only component of the monad.
 data RO = RO
   { iRange    :: Range                  -- ^ Source code being analysed
-  , iVars     :: Map QName VarType      -- ^ Type of variable that are in scope
+  , iVars     :: Map Name VarType      -- ^ Type of variable that are in scope
 
   {- NOTE: We assume no shadowing between these two, so it does not matter
   where we look first. Similarly, we assume no shadowing with
   the existential type variable (in RW).  See `checkTShadowing`. -}
 
   , iTVars    :: [TParam]                  -- ^ Type variable that are in scope
-  , iTSyns    :: Map QName (DefLoc, TySyn) -- ^ Type synonyms that are in scope
-  , iNewtypes :: Map QName (DefLoc, Newtype)
+  , iTSyns    :: Map Name (DefLoc, TySyn) -- ^ Type synonyms that are in scope
+  , iNewtypes :: Map Name (DefLoc, Newtype)
    -- ^ Newtype declarations in scope
    --
    -- NOTE: type synonyms take precedence over newtype.  The reason is
@@ -156,6 +187,10 @@ data RO = RO
     -- ^ When this flag is set to true, bindings that lack signatures
     -- in where-blocks will never be generalized. Bindings with type
     -- signatures, and all bindings at top level are unaffected.
+
+  , iSolver :: CrySAT.Solver
+
+  , iPrimNames :: !PrimMap
   }
 
 -- | Read-write component of the monad.
@@ -164,7 +199,7 @@ data RW = RW
   , iWarnings :: ![(Range,Warning)]     -- ^ Collected warnings
   , iSubst    :: !Subst                 -- ^ Accumulated substitution
 
-  , iExistTVars  :: [Map QName Type]
+  , iExistTVars  :: [Map Name Type]
     -- ^ These keeps track of what existential type variables are available.
     -- When we start checking a function, we push a new scope for
     -- its arguments, and we pop it when we are done checking the function
@@ -205,6 +240,9 @@ instance Monad InferM where
 instance MonadFix InferM where
   mfix f        = IM (mfix (unIM . f))
 
+instance FreshM InferM where
+  liftSupply f = IM (liftSupply f)
+
 
 io :: IO a -> InferM a
 io m = IM $ inBase m
@@ -233,6 +271,18 @@ recordWarning w =
   do r <- curRange
      IM $ sets_ $ \s -> s { iWarnings = (r,w) : iWarnings s }
 
+getSolver :: InferM CrySAT.Solver
+getSolver =
+  do RO { .. } <- IM ask
+     return iSolver
+
+-- | Retrieve the mapping between identifiers and declarations in the prelude.
+getPrimMap :: InferM PrimMap
+getPrimMap  =
+  do RO { .. } <- IM ask
+     return iPrimNames
+
+
 --------------------------------------------------------------------------------
 newGoal :: ConstraintSource -> Prop -> InferM Goal
 newGoal goalSource goal =
@@ -337,7 +387,7 @@ newTVar' src extraBound k =
 
 
 -- | Generate a new free type variable.
-newTParam :: Maybe QName -> Kind -> InferM TParam
+newTParam :: Maybe Name -> Kind -> InferM TParam
 newTParam nm k = newName $ \s -> let x = seedTVar s
                                  in (TParam { tpUnique = x
                                             , tpKind   = k
@@ -412,7 +462,7 @@ varsWithAsmps =
 
 
 -- | Lookup the type of a variable.
-lookupVar :: QName -> InferM VarType
+lookupVar :: Name -> InferM VarType
 lookupVar x =
   do mb <- IM $ asks $ Map.lookup x . iVars
      case mb of
@@ -427,23 +477,23 @@ lookupVar x =
 
 -- | Lookup a type variable.  Return `Nothing` if there is no such variable
 -- in scope, in which case we must be dealing with a type constant.
-lookupTVar :: QName -> InferM (Maybe Type)
+lookupTVar :: Name -> InferM (Maybe Type)
 lookupTVar x = IM $ asks $ fmap (TVar . tpVar) . find this . iTVars
   where this tp = tpName tp == Just x
 
 -- | Lookup the definition of a type synonym.
-lookupTSyn :: QName -> InferM (Maybe TySyn)
+lookupTSyn :: Name -> InferM (Maybe TySyn)
 lookupTSyn x = fmap (fmap snd . Map.lookup x) getTSyns
 
 -- | Lookup the definition of a newtype
-lookupNewtype :: QName -> InferM (Maybe Newtype)
+lookupNewtype :: Name -> InferM (Maybe Newtype)
 lookupNewtype x = fmap (fmap snd . Map.lookup x) getNewtypes
 
 -- | Check if we already have a name for this existential type variable and,
 -- if so, return the definition.  If not, try to create a new definition,
 -- if this is allowed.  If not, returns nothing.
 
-existVar :: QName -> Kind -> InferM Type
+existVar :: Name -> Kind -> InferM Type
 existVar x k =
   do scopes <- iExistTVars <$> IM get
      case msum (map (Map.lookup x) scopes) of
@@ -464,15 +514,15 @@ existVar x k =
 
 
 -- | Returns the type synonyms that are currently in scope.
-getTSyns :: InferM (Map QName (DefLoc,TySyn))
+getTSyns :: InferM (Map Name (DefLoc,TySyn))
 getTSyns = IM $ asks iTSyns
 
 -- | Returns the newtype declarations that are in scope.
-getNewtypes :: InferM (Map QName (DefLoc,Newtype))
+getNewtypes :: InferM (Map Name (DefLoc,Newtype))
 getNewtypes = IM $ asks iNewtypes
 
 -- | Get the set of bound type variables that are in scope.
-getTVars :: InferM (Set QName)
+getTVars :: InferM (Set Name)
 getTVars = IM $ asks $ Set.fromList . mapMaybe tpName . iTVars
 
 -- | Return the keys of the bound variables that are in scope.
@@ -488,7 +538,7 @@ because it is confusing.  As a bonus, in the implementation we don't
 need to worry about where we lookup things (i.e., in the variable or
 type synonym environment. -}
 
-checkTShadowing :: String -> QName -> InferM ()
+checkTShadowing :: String -> Name -> InferM ()
 checkTShadowing this new =
   do ro <- IM ask
      rw <- IM get
@@ -537,28 +587,28 @@ withNewtype t (IM m) =
                                                      (iNewtypes r) }) m
 
 -- | The sub-computation is performed with the given variable in scope.
-withVarType :: QName -> VarType -> InferM a -> InferM a
+withVarType :: Name -> VarType -> InferM a -> InferM a
 withVarType x s (IM m) =
   IM $ mapReader (\r -> r { iVars = Map.insert x s (iVars r) }) m
 
-withVarTypes :: [(QName,VarType)] -> InferM a -> InferM a
+withVarTypes :: [(Name,VarType)] -> InferM a -> InferM a
 withVarTypes xs m = foldr (uncurry withVarType) m xs
 
-withVar :: QName -> Schema -> InferM a -> InferM a
+withVar :: Name -> Schema -> InferM a -> InferM a
 withVar x s = withVarType x (ExtVar s)
 
 
 -- | The sub-computation is performed with the given variables in scope.
-withMonoType :: (QName,Located Type) -> InferM a -> InferM a
+withMonoType :: (Name,Located Type) -> InferM a -> InferM a
 withMonoType (x,lt) = withVar x (Forall [] [] (thing lt))
 
 -- | The sub-computation is performed with the given variables in scope.
-withMonoTypes :: Map QName (Located Type) -> InferM a -> InferM a
+withMonoTypes :: Map Name (Located Type) -> InferM a -> InferM a
 withMonoTypes xs m = foldr withMonoType m (Map.toList xs)
 
 -- | The sub-computation is performed with the given type synonyms
 -- and variables in scope.
-withDecls :: ([TySyn], Map QName Schema) -> InferM a -> InferM a
+withDecls :: ([TySyn], Map Name Schema) -> InferM a -> InferM a
 withDecls (ts,vs) m = foldr withTySyn (foldr add m (Map.toList vs)) ts
   where
   add (x,t) = withVar x t
@@ -581,11 +631,11 @@ inNewScope m =
 
 newtype KindM a = KM { unKM :: ReaderT KRO (StateT KRW InferM)  a }
 
-data KRO = KRO { lazyTVars  :: Map QName Type -- ^ lazy map, with tyvars.
+data KRO = KRO { lazyTVars  :: Map Name Type -- ^ lazy map, with tyvars.
                , allowWild  :: Bool           -- ^ are type-wild cards allowed?
                }
 
-data KRW = KRW { typeParams :: Map QName Kind -- ^ kinds of (known) vars.
+data KRW = KRW { typeParams :: Map Name Kind -- ^ kinds of (known) vars.
                }
 
 instance Functor KindM where
@@ -611,8 +661,8 @@ in the process of computing.
 As a result we return the value of the sub-computation and the computed
 kinds of the type parameters. -}
 runKindM :: Bool                          -- Are type-wild cards allowed?
-         -> [(QName, Maybe Kind, Type)]   -- ^ See comment
-         -> KindM a -> InferM (a, Map QName Kind)
+         -> [(Name, Maybe Kind, Type)]   -- ^ See comment
+         -> KindM a -> InferM (a, Map Name Kind)
 runKindM wildOK vs (KM m) =
   do (a,kw) <- runStateT krw (runReaderT kro m)
      return (a, typeParams kw)
@@ -626,7 +676,7 @@ data LkpTyVar = TLocalVar Type (Maybe Kind) -- ^ Locally bound variable.
               | TOuterVar Type              -- ^ An outer binding.
 
 -- | Check if a name refers to a type variable.
-kLookupTyVar :: QName -> KindM (Maybe LkpTyVar)
+kLookupTyVar :: Name -> KindM (Maybe LkpTyVar)
 kLookupTyVar x = KM $
   do vs <- lazyTVars `fmap` ask
      ss <- get
@@ -654,14 +704,14 @@ kNewType src k =
      kInInferM $ TVar `fmap` newTVar' src tps k
 
 -- | Lookup the definition of a type synonym.
-kLookupTSyn :: QName -> KindM (Maybe TySyn)
+kLookupTSyn :: Name -> KindM (Maybe TySyn)
 kLookupTSyn x = kInInferM $ lookupTSyn x
 
 -- | Lookup the definition of a newtype.
-kLookupNewtype :: QName -> KindM (Maybe Newtype)
+kLookupNewtype :: Name -> KindM (Maybe Newtype)
 kLookupNewtype x = kInInferM $ lookupNewtype x
 
-kExistTVar :: QName -> Kind -> KindM Type
+kExistTVar :: Name -> Kind -> KindM Type
 kExistTVar x k = kInInferM $ existVar x k
 
 -- | Replace the given bound variables with concrete types.
@@ -672,7 +722,7 @@ kInstantiateT t as = return (apSubst su t)
 {- | Record the kind for a local type variable.
 This assumes that we already checked that there was no other valid
 kind for the variable (if there was one, it gets over-written). -}
-kSetKind :: QName -> Kind -> KindM ()
+kSetKind :: Name -> Kind -> KindM ()
 kSetKind v k = KM $ sets_ $ \s -> s{ typeParams = Map.insert v k (typeParams s)}
 
 -- | The sub-computation is about the given range of the source code.
