@@ -1,13 +1,11 @@
 -- |
 -- Module      :  $Header$
--- Copyright   :  (c) 2013-2015 Galois, Inc.
+-- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
 -- Stability   :  provisional
 -- Portability :  portable
-
-{-# LANGUAGE PatternGuards #-}
-
+{-# LANGUAGE PatternGuards, Safe #-}
 module Cryptol.TypeCheck.Solver.Selector (tryHasGoal) where
 
 import Cryptol.TypeCheck.AST
@@ -16,13 +14,14 @@ import Cryptol.TypeCheck.Monad( InferM, unify, newGoals, lookupNewtype
                               , newType, applySubst, addHasGoal, solveHasGoal
                               )
 import Cryptol.TypeCheck.Subst(listSubst,apSubst)
+import Cryptol.Utils.Ident (Ident)
 import Cryptol.Utils.PP(text,pp,ordinal,(<+>))
 import Cryptol.Utils.Panic(panic)
 
 import Control.Monad(forM,guard)
 
 
-recordType :: [Name] -> InferM Type
+recordType :: [Ident] -> InferM Type
 recordType labels =
   do fields <- forM labels $ \l ->
         do t <- newType (text "record field" <+> pp l) KType
